#include <Adafruit_GFX.h>
#include <SPI.h>
#include <Wire.h>
#include <Adafruit_ILI9341.h>
#include <Adafruit_TSC2007.h>
#include "Math.h"

// function declarations
boolean checkCollision(int x1, int y1, int width1, int height1, int x2, int y2, int width2, int height2);

// For the Adafruit shield, these are the default.
#define TS_MINX 150
#define TS_MINY 130
#define TS_MAXX 3800
#define TS_MAXY 4000
#define TS_MIN_PRESSURE 100
Adafruit_TSC2007 ts;
// The display also uses hardware SPI, plus #9 & #10
#define TFT_CS 10
#define TFT_DC 9
Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS, TFT_DC);

#define BAT_WIDTH 40
#define BAT_HEIGHT 10
int batX, batY;


float xCalM = 0.0, yCalM = 0.0;
float xCalC = 0.0, yCalC = 0.0;
float xPos = 0.0, yPos = 0.0;
float xPosLast = 0.0, yPosLast = 0.0;
float xVel = 0.0, yVel = 0.0;
int8_t topBorder = 20;
int8_t ballSize = 3;
int playerLives = 3;
int playerScore = 0;
int gameState = 1;  // 1=start 2=playing 3=gameover
int level;

int16_t tftWidth = 0, tftHeight = 0;

class ScreenPoint {
public:
  int16_t x;
  int16_t y;

  ScreenPoint() {
    // default contructor
  }

  ScreenPoint(int16_t xIn, int16_t yIn) {
    x = xIn;
    y = yIn;
  }
};

class Block {
public:
  int x;
  int y;
  int width;
  int height;
  int colour;
  int score;
  boolean isActive;

  // default constructor
  Block() {}

  Block(int xpos, int ypos, int bwidth, int bheight, int bcol, int bscore) {
    x = xpos;
    y = ypos;
    width = bwidth;
    height = bheight;
    colour = bcol;
    score = bscore;
    isActive = true;
    drawBlock();
  }

  void drawBlock() {
    tft.fillRect(x, y, width, height, colour);
  }

  void removeBlock() {
    tft.fillRect(x, y, width, height, ILI9341_BLACK);
    isActive = false;
  }

  boolean isHit(float x1, float y1, int w1, int h1) {
    if (checkCollision((int)round(x1), (int)round(y1), w1, h1, x, y, width, height)) {
      return true;
    } else {
      return false;
    }
  }
};
Block blocks[5][16];

ScreenPoint getScreenCoords(TS_Point touchPoint) {
  uint16_t rawX, rawY;
  ScreenPoint screenPoint;
  screenPoint.x = map(rawX, TS_MINX, TS_MAXX, 0, tft.width());
  screenPoint.y = map(rawY, TS_MINY, TS_MAXY, 0, tft.height());
  return screenPoint;
}





void setup() {
  Serial.begin(115200);
  Serial.println("Startup");

  tft.begin();
  if (!ts.begin()) {
    Serial.println("Couldn't start touchscreen controller");
    while (1)
      ;
  }
  Serial.println("Touchscreen started");
  tft.setRotation(1);
  tft.fillScreen(ILI9341_BLACK);
  tftWidth = tft.width();
  tftHeight = tft.height();
  ts.begin();

  batX = (tft.width() - BAT_WIDTH) / 2;
  batY = tft.height() - 20;  // Adjust this value according to your preference
}


void initGame() {
  // Clear the screen and draw game elements
  tft.fillScreen(ILI9341_BLACK);
  tft.drawFastHLine(0, topBorder - 1, tftWidth, ILI9341_BLUE);

  // Display initial player information: score, lives, and level
  drawScore();
  drawLives();
  drawLevel();

  // Reset player lives, score, and level
  playerLives = 3;
  playerScore = 0;
  level = 0;

  // Initialize the game board with blocks
  initGameBoard();

  // Reset ball position and velocity
  newBall();

  // Set the game state to 'play'
  gameState = 2;
}



void initGameBoard() {
  int row, col;
  int colour, score;
  clearOldBallPos();
  xPosLast = xPos = 0;
  yPosLast = yPos = 90;
  xVel = 2;
  yVel = 2 + (level);
  gameState = 2;

  for (row = 0; row < 5; row++) {
    for (col = 0; col < 16; col++) {
      switch (row) {
        case 0:
        case 1:
          colour = ILI9341_BLUE;
          score = 50;
          break;
        case 2:
        case 3:
          colour = ILI9341_MAGENTA;
          score = 30;
          break;
        case 4:
        case 5:
          colour = ILI9341_YELLOW;
          score = 10;
          break;
      }
      // Use calibrated screen coordinates for block positions
      blocks[row][col] = Block(col * 20, (row * 10) + 30, 19, 9, colour, score);
    }
  }
}


void clearOldBallPos() {
  tft.fillCircle(round(xPosLast), round(yPosLast), ballSize, ILI9341_BLACK);
}

void moveBall() {
  float newX, newY;
  newX = xPos + xVel;
  newY = yPos + yVel;
  if (newX < (float)ballSize) {
    newX = (float)ballSize;
    xVel = -xVel;
  }
  if (newX > (float)(tftWidth - ballSize - 1)) {
    newX = (float)(tftWidth - ballSize - 1);
    xVel = -xVel;
  }
  if (newY < topBorder + (float)ballSize) {
    newY = topBorder + (float)ballSize;
    yVel = -yVel;
  }
  if ((round(newX) != round(xPosLast)) || (round(newY) != round(yPosLast))) {
    // Use calibrated screen coordinates for drawing ball
    clearOldBallPos();
    tft.fillCircle(round(newX), round(newY), ballSize, ILI9341_GREEN);
    xPosLast = newX;
    yPosLast = newY;
  }
  xPos = newX;
  yPos = newY;
}


void drawScore() {
  // clear old score
  tft.fillRect(50, 5, 25, 10, ILI9341_BLACK);
  // print new score
  tft.setCursor(50, 5);
  tft.setTextSize(1);
  tft.setTextColor(ILI9341_WHITE);
  tft.print(playerScore);

  if (playerScore == 0) {
    tft.print("Score: 0");
  } else {
    tft.print("Score: ");
    tft.print(playerScore);
  }
}

void drawLives() {
  // clear old lives
  tft.fillRect((tftWidth / 2) + 50, 5, 25, 10, ILI9341_BLACK);
  // print new score
  tft.setCursor((tftWidth / 2) + 50, 5);
  tft.setTextSize(1);
  tft.setTextColor(ILI9341_WHITE);
  tft.print(playerLives);
}

void drawLevel() {
  // clear old level
  tft.fillRect(tftWidth - 25, 5, 25, 10, ILI9341_BLACK);
  // print new score
  tft.setCursor(tftWidth - 25, 5);
  tft.setTextSize(1);
  tft.setTextColor(ILI9341_WHITE);
  tft.print(level + 1);
}

void newBall() {
  xPos = 0;
  yPos = 90;
  xVel = yVel = 2;
  moveBall();
  delay(1000);
}

boolean checkBallLost() {
  if (yPos > tftHeight + ballSize) {
    return true;
  } else {
    return false;
  }
}

boolean isScreenTouched() {
  uint16_t x, y, z1, z2;
  if (ts.read_touch(&x, &y, &z1, &z2) && (z1 > TS_MIN_PRESSURE)) {
    Serial.print("Touch point: (");
    Serial.print(x);
    Serial.print(", ");
    Serial.print(y);
    Serial.print(", ");
    Serial.print(z1);
    Serial.print(" / ");
    Serial.print(z2);
    Serial.println(")");
    x = map(x, TS_MINX, TS_MAXX, 0, tft.width());
    y = map(y, TS_MINY, TS_MAXY, 0, tft.height());
    TS_Point touchPoint(x, y, z1);  // Move the bat based on the touch point
    return true;
  }
  return false;
}

void moveBat(TS_Point touchPoint) {
  uint16_t x, y, z1, z2;
  // Convert raw touch coordinates to calibrated screen coordinates
  if (ts.read_touch(&x, &y, &z1, &z2) && (z1 > TS_MIN_PRESSURE)) {
    ScreenPoint sp = getScreenCoords(TS_Point(x, y, 0));



    // Clear the previous bat position
    tft.fillRect(batX, batY, BAT_WIDTH, BAT_HEIGHT, ILI9341_BLACK);

    batX = map(x, TS_MINX, TS_MAXX, 0, tft.width() - BAT_WIDTH);

    // Draw the bat in its new position
    tft.fillRect(batX, batY, BAT_WIDTH, BAT_HEIGHT, ILI9341_RED);

    Serial.print("Touch point: (");
    Serial.print(x);
    Serial.print(", ");
    Serial.print(y);
    Serial.print(", ");
    Serial.print(z1);
    Serial.print(" / ");
    Serial.print(z2);
    Serial.println(")");
  }
}





// bounding box collision detection
boolean checkCollision(int x1, int y1, int width1, int height1, int x2, int y2, int width2, int height2) {
  boolean hit = false;
  if (
    (((x2 + width2) >= x1) && (x2 <= (x1 + width1)))
    && (((y2 + height2) >= y1) && (y2 <= (y1 + height1)))) {
    hit = true;
  }

  return hit;
}


void checkHitBat() {
  // check bat and bottom half of ball
  float xInc;
  boolean hit = checkCollision(batX, batY, BAT_WIDTH, BAT_HEIGHT, (int)round(xPos) - ballSize, (int)round(yPos), ballSize * 2, ballSize);
  if (hit) {
    // reverse ball y direction but increase speed
    yVel += 0.05;
    if (yVel > 5) {
      yVel = 5;
    }
    yVel = -yVel;
    // rounded bounce
    xInc = (xPos - (float)(batX + (BAT_WIDTH / 2))) / (float)BAT_WIDTH;
    xVel += 6 * xInc;
    if (abs(xVel) > 6) {
      if (xVel < 0) {
        xVel = -6;
      } else {
        xVel = 6;
      }
    }
    // make sure ball not hitting bat
    yPos = (float)(batY - ballSize - 1);
  }
}

void checkHitBlock() {
  int row, col;
  for (row = 0; row < 5; row++) {
    for (col = 0; col < 16; col++) {
      if (blocks[row][col].isActive && blocks[row][col].isHit(xPos, yPos, ballSize * 2, ballSize * 2)) {
        blocks[row][col].removeBlock();
        playerScore += blocks[row][col].score;
        drawScore();
        yVel = -yVel;
        return;
      }
    }
  }
}

boolean checkAllBlocksHit() {
  int row, col, actives;
  actives = 0;
  for (row = 0; row < 5; row++) {
    for (col = 0; col < 16; col++) {
      if (blocks[row][col].isActive) {
        return false;
      }
    }
  }

  return true;
}

unsigned long lastFrame = millis();

void loop() {

  // limit frame rate
  while ((millis() - lastFrame) < 20)
    ;
  lastFrame = millis();

  switch (gameState) {
    case 1:  // start
      gameState = 11;
      break;

    case 11:  // click to play
      tft.fillRect((tftWidth / 2) - 100, (tftHeight / 2) - 20, 200, 40, ILI9341_GREEN);
      tft.setCursor((tftWidth / 2) - 100 + 25, (tftHeight / 2) - 20 + 12);
      tft.setTextSize(2);
      tft.setTextColor(ILI9341_WHITE);
      tft.print("CLICK TO PLAY");
      gameState = 12;
      break;

    case 12:  // wait for click to play
      if (isScreenTouched()) {
        // If the screen is touched, get the touch point coordinates
        TS_Point p = ts.getPoint();

        // Convert the touch point coordinates to screen coordinates
        ScreenPoint sp = getScreenCoords(p);

        // Define the boundaries of the "click to play" button
        int buttonX = (tftWidth / 2) - 50;
        int buttonY = (tftHeight / 2) - 20;
        int buttonWidth = 100;
        int buttonHeight = 40;

        // Check if the touch point falls within the boundaries of the button
        initGame();
        gameState = 2;  // Start playing the game
        Serial.println("Has Game Started");
        if (checkCollision(sp.x, sp.y, 1, 100, buttonX, buttonY, buttonWidth, buttonHeight)) {

          // If the touch is within the button, initialize the game and start playing
        }
      }
      break;


    case 2:  // play
      // Read raw touch coordinates
      uint16_t rawX, rawY, z1, z2;
      bool touched = ts.read_touch(&rawX, &rawY, &z1, &z2);

      // Create a TS_Point object with rawx as x-coordinate
      TS_Point touchPoint(rawX, rawY, 0);  // Assuming rawY represents the y-coordinate

      // Move the bat based on the touch point. Modify this function call according to your logic.
      moveBat(touchPoint);

      // Other game logic...
      moveBall();
      checkHitBat();
      checkHitBlock();

      // Check if the ball is lost and update game state accordingly
      if (checkBallLost()) {
        playerLives--;
        drawLives();
        if (playerLives > 0) {
          newBall();
        } else {
          gameState = 3;  // end game
        }
      }

      // Check if all blocks are hit and update game state accordingly
      if (checkAllBlocksHit()) {
        gameState = 4;
      }
      break;



    case 4:  // new blocks
      delay(1000);
      level++;
      drawLevel();
      initGameBoard();
      break;

    case 3:  // end
      tft.fillScreen(ILI9341_BLACK);
      tft.setCursor((tftWidth / 2) - 150, 50);
      tft.setTextSize(3);
      tft.setTextColor(ILI9341_WHITE);
      tft.print("You Scored ");
      tft.print(playerScore);
      gameState = 11;  // click to play
      break;
  }
}
